/***************************************************************************
*  Media Downloader. Search and download as many images or videos that     *
*    fit your search criteria                                              *
*  Copyright (C) 2009-2011  Marco Bavagnoli - lil.deimos@gmail.com         *
*                                                                          *
*  This program is free software: you can redistribute it and/or modify    *
*  it under the terms of the GNU General Public License as published by    *
*  the Free Software Foundation, either version 3 of the License, or       *
*  (at your option) any later version.                                     *
*                                                                          *
*  This program is distributed in the hope that it will be useful,         *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*  GNU General Public License for more details.                            *
*                                                                          *
*  You should have received a copy of the GNU General Public License       *
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
***************************************************************************/

// inspired by http://doc.qt.nokia.com/latest/network-googlesuggest.html

#include <QObject>
#include <QXmlStreamReader>
#include <QHeaderView>
#include <QEvent>
#include <QKeyEvent>
#include <QDebug>
#include "googlesuggest.h"

googleSuggest::googleSuggest(QWidget *parent) :
    QLineEdit(parent)
{
    setCanSuggest(false);

    net = new network();
    connect(net, SIGNAL(downloadFinished(int,QByteArray)), this, SLOT(gotSuggest(int,QByteArray)) );

    timer = new QTimer(this);
    timer->setSingleShot(true);
    timer->setInterval(600);
    connect(timer, SIGNAL(timeout()), SLOT(enableSuggest()));
    connect(this, SIGNAL(textEdited(QString)), timer, SLOT(start()));

    listSuggested = new QTreeWidget(this);
    listSuggested->setStyleSheet("background-color: rgb(255,255,255);"
                                 "alternate-background-color: rgb(245,245,255);"
                                 "border: 15px solid qlineargradient(spread:pad, x1:0, y1:0, x2:1, y2:0, stop:0 rgb(255, 255, 255), stop:1 rgba(0, 60, 255));"
                                 "border-left: 15px solid qlineargradient(spread:pad, x1:0, y1:0, x2:1, y2:0, stop:0 rgb(255, 255, 255), stop:1 rgba(0, 60, 255));"
                                 "border-right: 15px solid qlineargradient(spread:pad, x1:0, y1:0, x2:1, y2:0, stop:1 rgb(255, 255, 255), stop:0 rgba(0, 60, 255));"
                                 "border-top-color: transparent;"
                                 "border-bottom-color: transparent;"
                                 );
    listSuggested->setColumnCount(2);
    listSuggested->setUniformRowHeights(true);
    listSuggested->setRootIsDecorated(false);
    listSuggested->setAlternatingRowColors(true);
    listSuggested->setEditTriggers(QTreeWidget::NoEditTriggers);
    listSuggested->setSelectionBehavior(QTreeWidget::SelectRows);
    listSuggested->setFrameStyle(QFrame::Box | QFrame::Plain);
    listSuggested->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    listSuggested->header()->hide();

    listSuggested->setMouseTracking(true);
    listSuggested->setFocusProxy(parent);
    listSuggested->setFocusPolicy(Qt::NoFocus);
    listSuggested->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    listSuggested->setWindowFlags(Qt::Popup);

    listSuggested->installEventFilter(this);

    connect(listSuggested, SIGNAL(itemClicked(QTreeWidgetItem*,int)),
            SLOT(suggestChoosen(QTreeWidgetItem*,int)));
}

googleSuggest::~googleSuggest()
{
    net->deleteLater();
    delete timer;
    delete listSuggested;
    delete horizontalLayout;
}

void googleSuggest::enableSuggest()
{
    if (timer->isActive()) return;
    canSuggest = true;
    findSuggest();
}

void googleSuggest::setCanSuggest( bool can )
{
    canSuggest = can;
}

void googleSuggest::findSuggest()
{
    QString locale = QLocale::system().name().replace("_", "-");
    if (locale.length() < 2)
        locale = "en-US";

    QString url = "http://suggestqueries.google.com/complete/search?ds=yt&output=toolbar&hl="
                  +locale+"&q="+this->text();
    net->get(url);
}

void googleSuggest::gotSuggest(int ID, QByteArray data)
{
    Q_UNUSED(ID)
    QStringList results, hits;

    QXmlStreamReader xml(data);
    while (!xml.atEnd()) {
        xml.readNext();
        if (xml.tokenType() == QXmlStreamReader::StartElement) {
            if (xml.name() == "suggestion") {
                QStringRef str = xml.attributes().value("data");
                results << str.toString();
            }
            if (xml.name() == "num_queries") {
                QStringRef str = xml.attributes().value("int");
                hits << str.toString();
            }
        }
    }

    listSuggested->clear();
    for (int i = 0; i < results.count(); ++i) {
        QTreeWidgetItem * item;
        item = new QTreeWidgetItem(listSuggested);
        item->setText(0, results[i]);
        item->setText(1, hits[i]);
        item->setTextAlignment(1, Qt::AlignRight);
        item->setTextColor(1, Qt::gray);
    }
    listSuggested->resizeColumnToContents(0);
    listSuggested->resizeColumnToContents(1);
    listSuggested->adjustSize();
    listSuggested->setUpdatesEnabled(true);

    int h = listSuggested->sizeHintForRow(0) * results.count() + 30;
    listSuggested->resize(listSuggested->width()+30, h);


    listSuggested->move( qobject_cast<QWidget *>(this->parent())->mapToGlobal( QPoint( this->x(), this->y()+this->height()) ) );
    if (results.count()>0) {
        listSuggested->show();
        listSuggested->setFocus();
    }
}


void googleSuggest::suggestChoosen(QTreeWidgetItem *item ,int i)
{
    Q_UNUSED(i)
    setCanSuggest(false);
    this->setText(item->text(0));
    this->setFocus();
    listSuggested->hide();
}


bool googleSuggest::eventFilter(QObject * watched, QEvent * event)
{
    if (watched != listSuggested)
        return QObject::eventFilter(watched, event);

    if (event->type() == QEvent::MouseButtonPress) {
        setCanSuggest(false);
        listSuggested->hide();
        this->setFocus();
        return true;
    }

    if (event->type() == QEvent::KeyPress) {

        bool keyProcessed = false;

        int key = static_cast<QKeyEvent*>(event)->key();
        switch (key) {
        case Qt::Key_Enter:
        case Qt::Key_Return:
            setCanSuggest(false);
            qDebug() << "CLCIKEKED";
            if (listSuggested->currentItem()) {
                suggestChoosen(listSuggested->currentItem(), 0);
                keyProcessed = true;
                this->setFocus();
            }
            else {
                this->setFocus();
                this->event(event);
                listSuggested->hide();
            }
            break;

        case Qt::Key_Escape:
            this->setFocus();
            listSuggested->hide();
            keyProcessed = true;
            break;

        case Qt::Key_Up:
        case Qt::Key_Down:
        case Qt::Key_Home:
        case Qt::Key_End:
        case Qt::Key_PageUp:
        case Qt::Key_PageDown:
            break;

        default:
            this->setFocus();
            this->event(event);
            listSuggested->hide();
            break;
        }
        return keyProcessed;
    }

    return false;
}

