/***************************************************************************
*  Media Downloader. Search and download as many images or videos that     *
*    fit your search criteria                                              *
*  Copyright (C) 2009-2011  Marco Bavagnoli - lil.deimos@gmail.com         *
*                                                                          *
*  This program is free software: you can redistribute it and/or modify    *
*  it under the terms of the GNU General Public License as published by    *
*  the Free Software Foundation, either version 3 of the License, or       *
*  (at your option) any later version.                                     *
*                                                                          *
*  This program is distributed in the hope that it will be useful,         *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*  GNU General Public License for more details.                            *
*                                                                          *
*  You should have received a copy of the GNU General Public License       *
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
***************************************************************************/


#include <QApplication>
#include <QDesktopWidget>
#include <QDebug>
#include "slideshow.h"
#include "mainwindow.h"
#include "imagecontainer.h"

slideshow::slideshow(QList<imageContainer *> itemsList,
                     int _delay,
                     bool _fullScreen,
                     bool _loop,
                     bool _showAll,
                     bool _showImages,
                     bool _showVideos,
                     QWidget *parent) :
        QDialog(parent)
{
    items=itemsList;
    if (items.size()==0) return;

    timer_delay=delay=_delay;
    fullScreen=_fullScreen;
    loop=_loop;
    showAll=_showAll;
    showImages=_showImages;
    showVideos=_showVideos;


    timer = new QTimer(this);
    connect(timer, SIGNAL(timeout()), this, SLOT(playNextItem()));

    net=new network(false,MainWindow::pointerToMainWindow);
    connect(net, SIGNAL(downloadFinished(int,QByteArray)),
            this, SLOT(downloaded(int,QByteArray)));
    connect(net, SIGNAL(downloadError(int,QNetworkReply::NetworkError)),
            this, SLOT(downloadError(int,QNetworkReply::NetworkError)) );
    connect(net, SIGNAL(downloadProgress(int,qint64,qint64)), this, SLOT(downloadProgress(int,qint64,qint64)) );


    curr_item_index=-1;
    noMoreItems=false;

    player = new mediaPlayer(false, items.at(0)->image.kind,
                             MainWindow::pointerToMainWindow);
    connect(player, SIGNAL(keyPressed(QKeyEvent*)), this, SLOT(onKeyPressed(QKeyEvent*)) );
#ifdef PHONON_LIB
    connect(player, SIGNAL(videoPlayFinished()), this, SLOT(playNextItem()) );
#endif

    progress = new progressWidget(true, player);

    // Inizialize mediaplayer
    if (fullScreen)
        player->goFullScreen();
    else
        player->goNormalScreen();

    player->show();

    playNextItem();
}

slideshow::~slideshow()
{
    timer->stop();
    net->disconnect(this);
    net->on_PB_cancel_clicked();
    net->deleteLater();

    // TODO: try to stop singleShot with this ?? else crashes
    QTimer::singleShot(0,this, SLOT(hide()) );

    delete timer;
    if (!player.isNull()) delete player;
    if (!progress.isNull()) delete progress;
}

void slideshow::onKeyPressed( QKeyEvent * event )
{
    if (event->key() == Qt::Key_Escape) {this->close();this->deleteLater();}
    if (event->key() == Qt::Key_Left)  {
        net->cancellAllDownloads();
        progress->percent=0;
        timer->stop();
        playPreviousItem();
    }
    if (event->key() == Qt::Key_Right) {
        net->cancellAllDownloads();
        progress->percent=0;
        timer->stop();
        playNextItem();
    }
    if (event->key() == Qt::Key_Space) {
        if (timer_delay==0)
            timer_delay=delay;
        else {
            timer_delay=0;
            timer->stop();
        }
        if (timer_delay!=0) timer->start(timer_delay*1000);
    }

}

void slideshow::keyPressEvent( QKeyEvent * event )
{
    Q_UNUSED(event)
    // nothing to do. The event is catched by onKeyPressed emitted by player
}

void slideshow::backwardCurrItem()
{
    if (curr_item_index-1>=0) curr_item_index--;
    else {
        if (loop) curr_item_index=items.size();
        else this->close();
    }
}

void slideshow::forwardCurrItem()
{
    bool foundNext=false;
    noMoreItems=false;
    int ID=curr_item_index;

    // loop to find the next item to play
    do {
        if (curr_item_index+1<items.size()) curr_item_index++;
        else {
            if (loop) curr_item_index=0;
            else {
                // If timer is on and this is the last item, close after user interval
                // if displaying an image; if its a video close immediiately
                if (timer_delay>0)
                {
                    timer->stop();
                    noMoreItems=true;
                    foundNext=true;
                    if (items.at(curr_item_index)->image.kind & CONTAINER_KIND_IMAGE)
                        QTimer::singleShot(timer_delay*1000,this, SLOT(close()));
                    if (items.at(curr_item_index)->image.kind & CONTAINER_KIND_VIDEO)
                        this->close();
                    return;
                }
                else{
                    timer->stop();
                    noMoreItems=true;
                    foundNext=true;
                    this->close();
                }
            }
        }
        if (showImages) {
            if (items.at(curr_item_index)->image.kind & CONTAINER_KIND_IMAGE) foundNext=true;
        }
        if (showVideos) {
            if (items.at(curr_item_index)->image.kind & CONTAINER_KIND_VIDEO) foundNext=true;
        }
        if (showAll) foundNext=true;
        if (curr_item_index==ID) {
            noMoreItems=true;
            curr_item_index=ID;
        }
    } while (!foundNext && !noMoreItems);
}

void slideshow::playPreviousItem()
{
    backwardCurrItem();
    loadCurrItem();
}

void slideshow::playNextItem()
{
    forwardCurrItem();
    loadCurrItem();
}

void slideshow::loadCurrItem()
{
    if (noMoreItems) return;
//    qDebug() << "curr_item_index " << curr_item_index;
    if (items.at(curr_item_index)->image.kind & CONTAINER_KIND_IMAGE)
    {
        player->show();
        if (fullScreen)
            progress->setPosition(QApplication::desktop()->screenGeometry(-1).width() / 2,
                                  QApplication::desktop()->screenGeometry(-1).height() / 2);
        else
            progress->setPosition(player->geometry().width()/2, player->geometry().height()/2);
        progress->percent=0;
        progress->setProgressColor(0, 0, 255);
        progress->show();
        net->get(items.at(curr_item_index)->image.image_url);
    }
    if (items.at(curr_item_index)->image.kind & CONTAINER_KIND_VIDEO)
        displayCurrItem();
}

void slideshow::downloadProgress(int i, qint64 current, qint64 total)
{
    Q_UNUSED(i);
    if (items.at(curr_item_index)->image.kind & CONTAINER_KIND_IMAGE) {
        progress->setProgress(current,total);
    }
}

void slideshow::downloaded(int n, QByteArray buffer)
{
    Q_UNUSED(n);
    if (!buffer.isEmpty() && pixmap.loadFromData(buffer))
        imageLoaded=true;
    else {
            imageLoaded=false;
            downloadError(0, QNetworkReply::OperationCanceledError);
            return;
    }
    progress->percent=0;
    progress->hide();
    displayCurrItem();
}

void slideshow::downloadError(int i, QNetworkReply::NetworkError e)
{
    Q_UNUSED(e);
    Q_UNUSED(i);

    // left, right or esc key pressed
    qDebug() << e;
    if (e==QNetworkReply::OperationCanceledError) return;

    QPainter painter;
    QPixmap p(MainWindow::pointerToMainWindow->size());

    progress->hide();
    qDebug() << "Error: " << e;

    // Display an image with a big red X
    painter.begin(&p);
    painter.setRenderHint(QPainter::Antialiasing);
    painter.eraseRect(0,0,MainWindow::pointerToMainWindow->width(), MainWindow::pointerToMainWindow->height());
    painter.fillRect(0,0,MainWindow::pointerToMainWindow->width(), MainWindow::pointerToMainWindow->height(),QBrush(QColor(220,220,220,0)));
    painter.setPen(QPen(Qt::red, MainWindow::pointerToMainWindow->width()/30, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin));
    painter.drawLine(0,0, MainWindow::pointerToMainWindow->width(), MainWindow::pointerToMainWindow->height() );
    painter.drawLine(0,MainWindow::pointerToMainWindow->height(), MainWindow::pointerToMainWindow->width(), 0 );
    painter.end();
    pixmap=p;

    imageLoaded=true;
    displayCurrItem();
}

void slideshow::displayCurrItem()
{
    timer->stop();

    player->kind=items.at(curr_item_index)->image.kind;
    if (items.at(curr_item_index)->image.kind & CONTAINER_KIND_IMAGE && !pixmap.isNull()) {
        //player->show();

        if (timer_delay>0 && !noMoreItems) {
            timer->start(timer_delay*1000);
            timer->setSingleShot(true);
        }
        if (imageLoaded) player->play(pixmap);
        imageLoaded=false;
    }
#ifdef PHONON_LIB
    if ((items.at(curr_item_index)->image.kind & CONTAINER_KIND_VIDEO) && !timer->isActive())
    {
        if (items.at(curr_item_index)->image.kind == CONTAINER_KIND_VIDEO_LOCAL)
        {
            player->show();
            player->play(items.at(curr_item_index)->image.image_url);
        } else
        {
            // Reads infos about the video if not already set
            if (items.at(curr_item_index)->image.youTubeInfo.video_info.size()==0)
                if (youTube::getTubeVideoInfo(items.at(curr_item_index)->image.youTubeInfo)==0) return;

            player->show();
            player->play(items.at(curr_item_index)->image.youTubeInfo);
        }
    }
#endif
}
