/***************************************************************************
*  Media Downloader. Search and download as many images or videos that     *
*    fit your search criteria                                              *
*  Copyright (C) 2009-2011  Marco Bavagnoli - lil.deimos@gmail.com         *
*                                                                          *
*  This program is free software: you can redistribute it and/or modify    *
*  it under the terms of the GNU General Public License as published by    *
*  the Free Software Foundation, either version 3 of the License, or       *
*  (at your option) any later version.                                     *
*                                                                          *
*  This program is distributed in the hope that it will be useful,         *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*  GNU General Public License for more details.                            *
*                                                                          *
*  You should have received a copy of the GNU General Public License       *
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
***************************************************************************/

#include <QDebug>
#include <QDomDocument>
#include <QList>
#include <QRegExp>
#include <QMessageBox>
#include "mainwindow.h"
#include "youtube.h"
#include "network.h"

QString unescapeCharacters(QString s)
{
    QRegExp rx;
    QChar c;
    int pos;
    bool ok;

    rx.setMinimal(true);
    rx.setPattern("(%[A-F|0-9].)");
    pos=0;
    while ((pos = rx.indexIn(s, pos)) != -1)
    {
        c=rx.cap(1).toInt(&ok, 16);
        s=s.replace(rx.cap(1), c);
    }
    return s;
}

const video_formats youTube::videoFormats[10] = {
{37,"1920x1080","MP4",1920*1080},
{22,"1280x720","MP4",1280*720},
{35,"854x480","FLV",854*480},
{34,"640x480","FLV",640*480},
{0,"640x480","FLV",640*480},
{5,"640x480","FLV",640*480},
{18,"480x360","MP4",480*360},
{6,"480x360","FLV",480*360},
{13,"176x144","3GP",176*144},
{17,"176x144","3GP",176*144},
};

int youTube::ID_preferred_video_format=0;
bool youTube::only_mp4=true;

youTube::youTube()
{
}

/////////////////////////////////////////////////////////////////////////////
// search into &tube an existing resolution less or
// equal than videoFormats[ID_preferred_video_format] and
// return the index of QList<youTubeVideoInfo> video_info to play or download
int youTube::getPreferredResolution(YouTubeStruct &tube, int ID)
{
    bool b=false;
    int i,n;
    int found=-1;

//    qDebug() << "Preferred ID " << ID << "(" << videoFormats[ID].res << ")";
    n=ID;
    while (!b && n<=9) {
        for (i=0; i<tube.video_info.size(); i++)
            if (videoFormats[n].fmt == tube.video_info.at(i).fmt.toInt())
            {
                if (only_mp4) {
                    if (videoFormats[n].container=="MP4") {
                        found=i;
                        b=true;
                    }
                }
                else {
                    found=i;
                    b=true;
                }
            }
        n++;
    }
    // If no match resolution try to get a lower one
    if (!b) {
        long delta=1920*1080;
        i=0;
        while (i<tube.video_info.size()) {
            if (abs(tube.video_info.at(i).xres*tube.video_info.at(i).yres - videoFormats[ID].area) < delta)
            {
                if (only_mp4) {
                    if (videoFormats[i].container=="MP4") {
                        delta=abs(tube.video_info.at(i).xres*tube.video_info.at(i).yres - videoFormats[ID].area);
                        found = i;
                    }
                }
                else {
                    delta=abs(tube.video_info.at(i).xres*tube.video_info.at(i).yres - videoFormats[ID].area);
                    found = i;
                }
            }
            i++;
        }
    }
    // I surrender. Give the 1st in the list
    if (found==-1) found = 0;
//    qDebug() << "Preferred res: " << tube.video_info.at(found).fmt.toInt() <<
//                " - " << tube.video_info.at(found).humanFmt <<
//                " - " << tube.video_info.at(found).container;
    return found;
}

/////////////////////////////////////////////////////////////////////////////
// Retrieve a list of video formats available to be played or downloaded
// from tube.id
int youTube::getTubeVideoInfo(YouTubeStruct &tube)
{
    if (tube.video_info.size()>0) return tube.video_info.size(); // If info are already filled return;
    youTubeVideoInfo mode;
    int count=0;

    // Retrieve video id from url (QString id)
    QRegExp rx;
    rx.setPattern(".*videos/(.*)");
    rx.setMinimal(false);
    rx.indexIn(tube.id);
    if (rx.numCaptures()<1) {
        QMessageBox::critical ( 0, "Error", "Error in the retrieved url:\n"+tube.id);
        return count;
    }
    QString videoID=rx.cap(1);

    // Gather video information and output error if fails
    network *net;
    QByteArray data,data_percented;
    net = new network(false, MainWindow::pointerToMainWindow);
    data=net->get_synchronous("http://www.youtube.com/get_video_info?&video_id="+videoID+"&el=embedded&ps=default&eurl=&gl=US&hl=en",false);
    // If fails retry with different &el value
    if (data.isEmpty() || data.startsWith("status=fail"))
        data=net->get_synchronous("http://www.youtube.com/get_video_info?&video_id="+videoID+"&el=detailpage&ps=default&eurl=&gl=US&hl=en",false);
    if (data.isEmpty() || data.startsWith("status=fail"))
        data=net->get_synchronous("http://www.youtube.com/get_video_info?&video_id="+videoID+"&el=vevo&ps=default&eurl=&gl=US&hl=en",false);

//    qDebug() << data;
    if (data.isEmpty()) {
        QMessageBox::critical ( 0, "Error", "Error gathering video information");
        return count;
    }
    if (data.startsWith("status=fail")) {
        rx.setPattern("reason=(.*)%");
        rx.setMinimal(true);
        rx.indexIn(data);
        QMessageBox::critical ( 0, "Error", "Error: "+rx.cap(1).replace("+"," "));
        return count;
    }
    delete net;

    data_percented=data;
    while (data.contains('%'))
        data = QByteArray().fromPercentEncoding(data);
//    qDebug() << data;

    // Retrieve video token
    rx.setPattern("&token=(.*)&");
    rx.setMinimal(true);
    rx.indexIn(data);
    if (rx.numCaptures()<1) {
        QMessageBox::critical ( 0, "Error", "Error gathering video token information");
        return count;
    }
    QString token=rx.cap(1);
    // See in human format
//    qDebug() << data.replace("&","\n&");

    // Retrieve available videos from &fmt_map argument
    QStringList l;
    int n;
    rx.setPattern("&fmt_map=(.*)&");
    rx.setMinimal(true);
    rx.indexIn(data);
    l=rx.cap(1).split(",");
    for (n=0; n<l.size(); n++)
    {
        mode.fmt=l.at(n).split("/").at(0);
        // http://en.wikipedia.org/wiki/YouTube#Quality_and_codecs
        switch (mode.fmt.toInt()) {
        case 0:
        case 5:
            mode.humanFmt="640x360";
            mode.container="FLV";
            mode.xres=640; mode.yres=360;
            break;
        case 6:
            mode.humanFmt="480x360";
            mode.container="FLV";
            mode.xres=480; mode.yres=360;
            break;
        case 13:
            mode.humanFmt="176x144";
            mode.container="3GP";
            mode.xres=480; mode.yres=360;
            break;
        case 17:
            mode.humanFmt="176x144";
            mode.container="3GP";
            mode.xres=176; mode.yres=144;
            break;
        case 18:
            mode.humanFmt="480x360";
            mode.container="MP4";
            mode.xres=480; mode.yres=360;
            break;
        case 22:
            mode.humanFmt="1280x720";
            mode.container="MP4";
            mode.xres=1280; mode.yres=720;
            break;
        case 34:
            mode.humanFmt="640x480";
            mode.container="FLV";
            mode.xres=640; mode.yres=480;
            break;
        case 35:
            mode.humanFmt="854x480";
            mode.container="FLV";
            mode.xres=854; mode.yres=480;
            break;
        case 37:
            mode.humanFmt="1920x1080";
            mode.container="MP4";
            mode.xres=1920; mode.yres=1080;
            break;
        case 38:
            mode.humanFmt="Original format";
            mode.container="MP4";
            mode.xres=0; mode.yres=0;
            break;
        default:
            mode.humanFmt="unknown mode:"+mode.fmt;
            mode.container="";
            mode.xres=320; mode.yres=240;
            break;
        }
        mode.url_play=getUrlsFromPattern(1, data_percented, mode.fmt.toInt());
        mode.url_retrieve=getUrlsFromPattern(0, data_percented, mode.fmt.toInt());
        // this works also but I think it's preferred to use the info data
//        mode.url_play="http://www.youtube.com/get_video?video_id="+videoID+"&t="+token+"&eurl=&el=&ps=&asv=&fmt="+mode.fmt;

//        qDebug() << data.replace("&","\n&");//.replace("&","\n&");
//        qDebug() << "mode.url_play" << mode.url_play;
//        qDebug() << "mode.url_retrieve" << mode.url_retrieve;

        qDebug() << "Video format found: " << mode.fmt << "-" << mode.humanFmt << " with " << mode.container << "format";
        count++;
        tube.video_info.append(mode);
    }
    tube.ID_preferred_video_format=youTube::getPreferredResolution(tube, youTube::ID_preferred_video_format);
    return count;
}

// Retrieve url from &fmt_stream_map
// These url are used for download or watching a video in different formats
// url_or_strem=0 search for fmt_strem_map
// url_or_strem=1 search for fmt_url_map
QString getUrlsFromPattern(int url_or_strem, QByteArray data, int itag)
{
    QString pattern;
    QString ret;
    QString fmt_map;
    QRegExp rx;
    int pos;

//    qDebug() << data;
    if (url_or_strem==0) pattern="&fmt_stream_map=(.*)&";
    if (url_or_strem==1) pattern="&fmt_url_map=(.*)&";
    rx.setPattern(pattern);
    rx.setMinimal(true);
    rx.indexIn(data);
    fmt_map=rx.cap(1);
    while (fmt_map.contains('%'))
        fmt_map = QByteArray().fromPercentEncoding(fmt_map.toAscii());

    fmt_map.append(",99|"); // Append this to intercept last url
//    qDebug() << fmt_map;

    if (url_or_strem==0) pattern="([0-9]{1,2})\\|(http://.*)\\|.*\\|.*,([0-9]{1,2}\\|)";
    if (url_or_strem==1) pattern="([0-9]{1,2})\\|(http://.*),([0-9]{1,2}\\|)";

    rx.setPattern(pattern);
    rx.setMinimal(true);

    pos=0;
    ret.clear();
    while ((pos = rx.indexIn(fmt_map, pos)) != -1)
    {
//        qDebug() << "CAP1 " << rx.cap(1);
//        qDebug() << "CAP2 " << rx.cap(2);
//        qDebug() << "CAP3 " << rx.cap(3);
        if (rx.cap(1)==QString::number(itag)) {// found url of the correct format
            ret=rx.cap(2);
            return ret;
        }
        pos += rx.matchedLength()-rx.cap(3).length();
    }
    return ret;
}

// Compose a string url with user parameters and
// fill tubes list with the retrieved videos
QList<YouTubeStruct> youTube::getList(QString search,
                                      StandardVideoFeeds feed,
                                      int start_feed_number, int feeds_to_get,
                                      int order,
                                      QString preferred_language,
                                      int time
                                      )
{
    QString getUrl;
    QString feedTime;
    QString orderBy;
    switch (time) {
        case 0: feedTime="all_time";   break;
        case 1: feedTime="today";      break;
        case 2: feedTime="this_week";  break;
        case 3: feedTime="this_month"; break;
    }
    switch (order) {
        case 0: orderBy="relevance";   break;
        case 1: orderBy="published";   break;
        case 2: orderBy="viewCount";   break;
        case 3: orderBy="rating";      break;
    }

    switch (feed) {
    case VIDEO_FEED_NONE_USE_SEARCH:
        getUrl="http://gdata.youtube.com/feeds/api/videos?q="+search.replace(" ","+");
        break;
    case VIDEO_FEED_TOP_RATED:
        getUrl="http://gdata.youtube.com/feeds/api/standardfeeds/top_rated";
        break;
    case VIDEO_FEED_TOP_FAVORITES:
        getUrl="http://gdata.youtube.com/feeds/api/standardfeeds/top_favorites";
        break;
    case VIDEO_FEED_MOST_VIEWED:
        getUrl="http://gdata.youtube.com/feeds/api/standardfeeds/most_viewed";
        break;
    case VIDEO_FEED_MOST_POPULAR:
        getUrl="http://gdata.youtube.com/feeds/api/standardfeeds/most_popular";
        break;
    case VIDEO_FEED_MOST_RECENT:
        getUrl="http://gdata.youtube.com/feeds/api/standardfeeds/most_recent";
        break;
    }
    // Only if we made a string search other parameters are added
    if (feed==VIDEO_FEED_NONE_USE_SEARCH)
    {
        getUrl+="&start-index="+QString::number(start_feed_number);
        getUrl+="&max-results="+QString::number(feeds_to_get);
        getUrl+="&orderby="+orderBy;
        if (!preferred_language.isEmpty()) getUrl+="&relevance_lang_languageCode="+preferred_language;
        getUrl+="&time="+feedTime;
        getUrl+="&v=1";
    }
    // Add this to getUrl to have a human output:  +"&prettyprint=true"
    return fillSearchTubes(getUrl);
}

// Make a specific search
QList<YouTubeStruct> youTube::fillSearchTubes(QString getUrl)
{
    QByteArray data;
    YouTubeStruct tube;
    QList<YouTubeStruct> tubes;

    network *net = new network(false);
    data=net->get_synchronous(getUrl,false);

//    qDebug() << data;
    net->deleteLater();

    if (data.isEmpty()) return tubes;

    QDomDocument doc("mydocument");
    doc.setContent(data);
    QDomElement docElem = doc.documentElement();

    QDomNode n = docElem.firstChild();
    QDomElement e,t;

    // Parse all <entry> tag to get infos about the videos
    tubes.clear();
    while(!n.isNull()) {
        e = n.toElement(); // try to convert the node to an element.
        if(!e.isNull()) {
            if (e.tagName()=="entry")
            {
                tube.id          = e.firstChildElement("id").text();
                tube.published   = e.firstChildElement("published").text();
                tube.updated     = e.firstChildElement("updated").text();
                tube.title       = e.firstChildElement("title").text();
                tube.content     = e.firstChildElement("content").text();
                tube.author_name = e.firstChildElement("author").firstChildElement("name").text();
                tube.author_uri  = e.firstChildElement("author").firstChildElement("uri").text();

                tube.duration    = e.firstChildElement("media:group").firstChildElement("yt:duration").attribute("seconds").toAscii();
                tube.mediaUrl    = e.firstChildElement("media:group").firstChildElement("media:player").attribute("url").toAscii().replace("&amp","&");
                t = e.firstChildElement("media:group").firstChildElement("media:thumbnail");
                tube.p[0].height = t.attribute("height").toInt();
                tube.p[0].width  = t.attribute("width").toInt();
                tube.p[0].url    = t.attribute("url").toAscii();
                t=t.nextSiblingElement("media:thumbnail");
                tube.p[1].height = t.attribute("height").toInt();
                tube.p[1].width  = t.attribute("width").toInt();
                tube.p[1].url    = t.attribute("url").toAscii();
                t=t.nextSiblingElement("media:thumbnail");
                tube.p[2].height = t.attribute("height").toInt();
                tube.p[2].width  = t.attribute("width").toInt();
                tube.p[2].url    = t.attribute("url").toAscii();
                t=t.nextSiblingElement("media:thumbnail");
                tube.p[3].height = t.attribute("height").toInt();
                tube.p[3].width  = t.attribute("width").toInt();
                tube.p[3].url    = t.attribute("url").toAscii();

                tube.rating_average       = e.firstChildElement("gd:rating").attribute("average").toAscii();
                tube.rating_numRaters     = e.firstChildElement("gd:rating").attribute("numRaters").toAscii();
                tube.statistics_viewCount = e.firstChildElement("yt:statistics").attribute("viewCount").toAscii();
                tube.video_id=tube.id.right(11);
                tubes.append(tube);
            }
        }
        n = n.nextSibling();
    }
    return tubes;
}

